import json
import os

import dash
import dash_html_components as html
import requests
from dash.dependencies import Input, Output, State
from sklearn.preprocessing import MultiLabelBinarizer

from . import io_utils, web_utils
from .mask_explain import MaskExplainer, plot_explanation
from .model import StandaloneModel
from .validation import dimension_reduction_plot, get_tag_states

URL = "https://gosheet-bqjlnzid4q-uc.a.run.app/add"


class Server:
    def __init__(self, state=None):
        self.dataset_p = "dataset.zip"
        self.model_p = "model"
        self.unlabelled_p = "data/unlabelled.json"
        self.state = state

        self.init_static()
        self.init_state()
        self.init_plot()

        self.case_str = ""
        self.tag_str = ""

        external_stylesheets = ["https://codepen.io/chriddyp/pen/bWLwgP.css"]
        self.app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

    def init_static(self):
        if not os.path.exists(self.dataset_p):
            io_utils.prepare_dataset(self.dataset_p)
        if not os.path.exists(self.model_p):
            io_utils.prepare_model(self.model_p)
        if not os.path.exists(self.unlabelled_p):
            io_utils.prepare_unlabelled(self.unlabelled_p)

    def init_state(self):
        self.rawdata = io_utils.load_datazip(self.dataset_p)
        self.unlabelled = io_utils.load_json(self.unlabelled_p)

        self.model = StandaloneModel.from_path(self.model_p)
        self.mlb = MultiLabelBinarizer().fit(self.rawdata.y_tags)
        self.mask_explainer = MaskExplainer(self.mlb)
        if self.state is None:
            self.state = get_tag_states(self.model, self.rawdata, self.mlb)

    def init_plot(self):
        self.fig, self.dimension_reducer = dimension_reduction_plot(
            self.state, method_n="tsne", n_components=2, dash=True
        )

    def plot(self):
        app = self.app
        input_id = "input"
        input_submit_id = "input_submit"
        dot_id = "t-sne"
        case_id = "cases"
        mask_id = "mask"
        checkbox_id = "checkbox"
        submit_idx = "submit"
        app.layout = html.Div(
            [
                web_utils.html_input(id_=input_id),
                web_utils.html_submit(id_=input_submit_id),
                html.H2("t-SNE"),
                web_utils.html_dots(dot_id, self.fig),
                html.H2("Explanation"),
                web_utils.html_case(id_=case_id),
                web_utils.html_mask(id_=mask_id),
                html.H2("Check the correct predictions"),
                web_utils.html_checkbox(id_=checkbox_id),
                web_utils.html_submit(id_=submit_idx),
            ]
        )

        @app.callback(
            [
                Output(case_id, "children"),
                Output(mask_id, "figure"),
                Output(mask_id, "style"),
                Output(checkbox_id, "options"),
                Output(checkbox_id, "value"),
                Output(submit_idx, "disabled"),
            ],
            [
                Input(input_submit_id, "n_clicks"),
                Input(dot_id, "clickData"),
            ],
            [State(input_id, "value")],
        )
        def update_output_div(n_clicks, clickData, input_value):
            if clickData is not None:
                return self._display_click_data(clickData)

            elif input_value is not None and n_clicks is not None:
                return self._case_plot({"COMMENT": input_value})

            return [[], web_utils.empty_bar(), {"display": "none"}, [], [], True]

        @app.callback(
            [
                Output(dot_id, "clickData"),
                Output(input_id, "value"),
                Output(input_submit_id, "n_clicks"),
            ],
            [Input(submit_idx, "n_clicks")],
            [State(checkbox_id, "value")],
        )
        def update_output(n_clicks, value):
            if n_clicks is None:
                raise dash.exceptions.PreventUpdate("cancel the callback")
            else:
                judge = {
                    "text": self.case_str,
                    "rets": self.tag_str,
                    "mistakes": json.dumps(value),
                }
                requests.post(URL, json=judge)

                return None, None, None

    def _display_click_data(self, clickData):

        customdata = clickData["points"][0]["customdata"]
        idx = customdata[0]
        from_ = customdata[1]
        pred_tag = customdata[2]
        if from_ == "unlabelled":
            case = self.unlabelled[idx]
        else:
            data = self.rawdata.retrive(from_, idx)
            case = data["text"]
        return self._case_plot(case, pred_tag)

    def _case_plot(self, case, pred_tag):
        childrend = []
        self.case_str = json.dumps(case)
        self.tag_str = json.dumps(pred_tag)
        prob = self.model.predict_prob([case], self.mlb)[0]

        options = [{"label": f"{c}: {p:.02f}", "value": c} for (c, p) in prob]
        if len(pred_tag) == 0:
            childrend.append(html.H3("No confidence in any predictions"))
            childrend.append(web_utils.dict_to_str(case))
            max_prob = max(prob, key=lambda x: x[1])
            values = [max_prob[0]]
            disabled_submit = False
            fig = web_utils.empty_bar()
            style = {"display": "none"}
        else:
            rets = self.mask_explainer.explain(self.model, case)
            values = [ret.tag for ret in rets]
            for ret in rets:
                importance = ret.importance
                pos_key_marks = [p[0] for p in importance][:5]
                childrend.append(html.H3(ret.tag))
                childrend.append(web_utils.draw_color(case, pos_key_marks))
            disabled_submit = False
            fig = plot_explanation(rets, dash=True)
            style = {"display": "block"}
        return [childrend, fig, style, options, values, disabled_submit]



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from torch import cuda, nn
from tqdm.autonotebook import tqdm
from transformers import AutoTokenizer, BertModel, BertPreTrainedModel
from transformers.modeling_outputs import SequenceClassifierOutput

from .data_utils import compose
from .domain import Mlb


class Classification(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.bert = BertModel.from_pretrained(config.identifier)
        self.dropout = nn.Dropout(config.dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        outputs = (logits,) + outputs[2:]

        loss = None
        if labels is not None:
            loss_fct = nn.BCEWithLogitsLoss()
            loss = loss_fct(
                logits.view(-1, self.num_labels), labels.view(-1, self.num_labels)
            )

        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )


class StandaloneModel:
    def __init__(self, model: Classification, tokenizer=None, max_len=200):
        self.model = model
        if tokenizer is None:
            tokenizer = get_tokenizer()
        self.tokenizer = tokenizer
        self.max_len = max_len
        self.device = "cuda" if cuda.is_available() else "cpu"
        self.model.to(self.device)

    @classmethod
    def from_path(cls, model_path: str, tokenizer=None, max_len=200):
        model = Classification.from_pretrained(model_path)
        return cls(model, tokenizer, max_len)

    def predict_tags(
        self, cases: list, mlb: Mlb, batch_size=8, tqdm_disable=True
    ) -> list:
        preds = self.predict(cases, batch_size=batch_size, tqdm_disable=tqdm_disable)
        return mlb.inverse_transform(preds >= 0.5)

    def predict_prob(
        self, cases: list, mlb: Mlb, batch_size=8, tqdm_disable=True
    ) -> List[List[Tuple[str, float]]]:
        preds = self.predict(cases, batch_size=batch_size, tqdm_disable=tqdm_disable)
        return [list(zip(mlb.classes_, pred)) for pred in preds]

    def predict(
        self, cases: list, batch_size=8, pooled_output=False, tqdm_disable=False
    ) -> np.array:
        """Predict the prob for 23 tags. If cases are dict, they will be
        composed by their `values` without shuffle.

        Args:
            cases (list): The input, can be either a list of str or a list of dict.
            batch_size (int, optional): The batch size. Defaults to 8.
            pooled_output (bool, optional): Whether set the output as 768-D vector. Defaults to False.

        Returns:
            np.array: [description]
        """
        pooled_input = False
        if isinstance(cases[0], dict):
            cases = list(map(compose, cases))
        elif isinstance(cases[0], np.array):
            pooled_input = True
        self.model.eval()
        preds: Optional[torch.Tensor] = None
        case_size = len(cases)
        for step in tqdm(range(0, case_size, batch_size), disable=tqdm_disable):
            batch = cases[step : step + batch_size]
            if pooled_input:
                outputs = self._predict_on_pooled_output(batch)
            else:
                inputs = self._encode(batch)
                outputs = self._predict_step(inputs, pooled_output=pooled_output)
            preds = outputs if preds is None else torch.cat((preds, outputs), dim=0)
        if preds is None:
            return preds
        if pooled_output:
            return preds.cpu().numpy()
        return torch.sigmoid(preds).cpu().numpy()

    def _encode(self, texts):
        return self.tokenizer.batch_encode_plus(
            texts,
            add_special_tokens=True,
            truncation=True,
            max_length=self.max_len,
            padding="max_length",
            return_token_type_ids=True,
        )

    def _predict_step(self, inputs, pooled_output=False):
        inputs = self._prepare_inputs(inputs)
        with torch.no_grad():
            if pooled_output:
                outputs = self.model.base_model(**inputs)
                pooled_output = outputs[1]
                return pooled_output.detach()
            outputs = self.model(**inputs)
            logits = outputs[0]
        return logits.detach()

    def _prepare_inputs(
        self, inputs: Dict[str, Union[torch.Tensor, Any]]
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Prepare :obj:`inputs` before feeding them to the model, converting them to tensors if they are not already and
        handling potential state.
        """
        for k, v in inputs.items():
            inputs[k] = torch.tensor(v, dtype=torch.long).to(
                self.device, dtype=torch.long
            )

        return inputs

    def _predict_on_pooled_output(self, pooled_output: np.array):
        with torch.no_grad():
            logits = self.model.classifier(pooled_output)
        return logits.detach()


def get_tokenizer():
    return AutoTokenizer.from_pretrained("bert-base-uncased")
